/*
 * This is the implementation of routines to get and set RAID information.
 *
 * Copyright 2003, Broadcom Corporation
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Broadcom Corporation;
 * the contents of this file may not be disclosed to third parties, copied
 * or duplicated in any form, in whole or in part, without the prior
 * written permission of Broadcom Corporation.
 *
 * $Id: print_server.c,v 1.2 2007/11/01 01:56:08 nasmaker Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <bcmnvram.h>
#include "notify_rc.h"
#include "shutils.h"
#include "print_server.h"
#include "disk_configuration.h"
#include "ipcmsg.h"
#include "typedefs.h"
#include <user_notification.h>
#include "translation/translation.h"

#include <linux/usb.h>

#define DEVICE_ID_LENGTH 1024
#define IOCNR_GET_DEVICE_ID 1
#define LPIOC_GET_DEVICE_ID(len) _IOC(_IOC_READ,'P', IOCNR_GET_DEVICE_ID,len)

#define MAN_DESC "MANUFACTURER:"
#define MAN_DESC2 "MFG:"
#define MDL_DESC "MODEL:"
#define MDL_DESC2 "MDL:"
#define SRL_DESC "SERIALNUMBER:"
#define SRL_DESC2 "SERN:"

#define UNKNOWN "N.A."
#define PRINTER "NASPrinter"

extern printer_info_t *get_active_printer_info()
{
  int fd,rc=0,major,minor,printer_no;
  struct stat stat_buf;
  unsigned char *device_id;
  struct printer_info_t *result;
  unsigned char *follow_info;
  unsigned char *start_info;
  int count=0;

  int serial_found=0;
  int manufacturer_found=0;
  int product_found=0;
  int id_len;

  char *serial;
  char *manufacturer;
  char *product;

  serial=NULL;
  manufacturer=NULL;
  product=NULL;

  result = NULL;

  for (printer_no=0;printer_no<MAX_PRINTERS;printer_no++)
    {
      char device_name[100];
      printer_desc_t *desc;
      printer_info_t *new_printer_info;
      printer_info_t **follow_printer_info_list;
      int unknown_no=1;

      sprintf(device_name,"/dev/usb/lp%d",printer_no);
      fd = open(device_name, O_RDONLY, 0);

      if (fd<0)
	{
	  //	  fprintf(stderr,"failed to open %s device\n",
	  //		  device_name);
	  continue;
	}

      rc = fstat(fd,  &stat_buf);
  
      if (rc)
	{
	  fprintf(stderr, "failed to query device %s\n",device_name);
	  close(fd);
	  continue;
	}

      if (!S_ISCHR(stat_buf.st_mode))
	{
	  fprintf(stderr,"%s is not a character device\n", device_name);
	  close(fd);
	  continue;
	}

      major = major(stat_buf.st_rdev);
      minor = minor(stat_buf.st_rdev);

//      printf("%s has a major of %d\n",device_name,major);
//      printf("%s has a minor of %d\n",device_name,minor);

      device_id = (unsigned char *) malloc(DEVICE_ID_LENGTH);
      
      if (device_id==NULL)
	{
	  fprintf(stderr,
		  "Failed trying to allocate %u bytes of memory"
		  "while querying device id for printers. \n",
		  DEVICE_ID_LENGTH);
	  close(fd);
	  continue;
	}

      rc = ioctl(fd,LPIOC_GET_DEVICE_ID(DEVICE_ID_LENGTH),device_id);

      follow_info=device_id;

      while (count<DEVICE_ID_LENGTH)
	{
	  if ((strncmp(follow_info,MAN_DESC,strlen(MAN_DESC))==0)||
	      (strncmp(follow_info,MAN_DESC2,strlen(MAN_DESC2))==0))

	    {   
	      int str_len;

	      /* MANUFACTURER_FOUND */
	      start_info = follow_info;

	      if (strncmp(follow_info,MAN_DESC,strlen(MAN_DESC))==0)
		str_len = strlen(MAN_DESC);
	      else
		str_len = strlen(MAN_DESC2);

	      while((count<DEVICE_ID_LENGTH) && (*follow_info!=';'))
		{
		  /* fill in white-space */
		  if (*follow_info==' ')
		    *follow_info='-';
		  ++follow_info;
		  ++count;
		}

	      if (count>=DEVICE_ID_LENGTH)
		{
		  if (serial_found)
		    free(serial);
		  if (manufacturer_found)
		    free(manufacturer);
		  if (product_found)
		    free(product);
		  close(fd);
		  return result;
		}

	      id_len = follow_info - start_info+1-str_len;
	      manufacturer = (char *) malloc(id_len);

	      strncpy(manufacturer,start_info+str_len,id_len-1);
	      manufacturer[id_len-1] = '\0';
	      manufacturer_found=1;
	    }

	  if ((strncmp(follow_info,MDL_DESC,strlen(MDL_DESC))==0)||
	      (strncmp(follow_info,MDL_DESC2,strlen(MDL_DESC2))==0))

	    {   
	      int str_len;

	      /* PRODUCT_MODEL_FOUND */
	      start_info = follow_info;

	      if (strncmp(follow_info,MDL_DESC,strlen(MDL_DESC))==0)
		str_len = strlen(MDL_DESC);
	      else
		str_len = strlen(MDL_DESC2);

	      while((count<DEVICE_ID_LENGTH) && (*follow_info!=';'))
		{
		  /* fill in white-space */
		  if (*follow_info==' ')
		    *follow_info='-';
		  ++follow_info;
		  ++count;
		}

	      if (count>=DEVICE_ID_LENGTH)
		{
		  if (serial_found)
		    free(serial);
		  if (manufacturer_found)
		    free(manufacturer);
		  if (product_found)
		    free(product);
		  close(fd);
		  return result;
		}

	      id_len = follow_info - start_info+1-str_len;

	      /* limit model name to 32 chars, imposed by samba's local.h via 
		 MAXPRINTERLEN */
	      if (id_len>32)
		id_len=32;

	      product = (char *) malloc(id_len);

	      strncpy(product,start_info+str_len,id_len-1);
	      product[id_len-1] = '\0';
	      product_found=1;
	    }

	  if ((strncmp(follow_info,SRL_DESC,strlen(SRL_DESC))==0)||
	      (strncmp(follow_info,SRL_DESC2,strlen(SRL_DESC2))==0))

	    {   
	      int str_len;

	      /* SERIAL_NUMBER_FOUND */
	      start_info = follow_info;

	      if (strncmp(follow_info,SRL_DESC,strlen(SRL_DESC))==0)
		str_len = strlen(SRL_DESC);
	      else
		str_len = strlen(SRL_DESC2);

	      while((count<DEVICE_ID_LENGTH) && (*follow_info!=';'))
		{
		  /* fill in white-space */
		  if (*follow_info==' ')
		    *follow_info='-';
		  ++follow_info;
		  ++count;
		}

	      if (count>=DEVICE_ID_LENGTH)
		{
		  if (serial_found)
		    free(serial);
		  if (manufacturer_found)
		    free(manufacturer);
		  if (product_found)
		    free(product);
		  close(fd);
		  return result;
		}

	      id_len = follow_info - start_info+1-str_len;
	      serial = (char *) malloc(id_len);

	      strncpy(serial,start_info+str_len,id_len-1);
	      serial[id_len-1] = '\0';
	      serial_found=1;
	    }
	  follow_info++;
	  count++;
	}

      new_printer_info = (printer_info_t *)(malloc(sizeof(printer_info_t)));
 
      if (new_printer_info==NULL)
       {
	 fprintf(stderr,
		 "Failed trying to allocate %lu bytes of memory"
		 "while querying usb for printers. \n",
		 (unsigned long) sizeof(printer_info_t));
	 
	 if (serial_found)
	   free(serial);
	 if (manufacturer_found)
	   free(manufacturer);
	 if (product_found)
	   free(product);
	 close(fd);
	 return result;
       }
     
     desc = (printer_desc_t *) (malloc(sizeof(printer_desc_t)));
     
     if (desc == NULL)
       {
	 fprintf(stderr,
		 "Failed trying to allocate %lu bytes of memory"
		 "while querying usb for printers. \n",
		 (unsigned long) sizeof(printer_desc_t));
	 
	 if (serial_found)
	   free(serial);
	 if (manufacturer_found)
	   free(manufacturer);
	 if (product_found)
	   free(product);
	 free(new_printer_info);
	 close(fd);
	 return result;
       }

     new_printer_info->desc = desc;

     if (serial_found)
       new_printer_info->desc->serial = serial;
     else
       {
	 int str_len;
	 str_len = strlen(UNKNOWN);
	 serial = malloc(str_len+1);
	 strncpy(serial,UNKNOWN,str_len);
	 serial[str_len]='\0';
	 new_printer_info->desc->serial = serial;
       }

     if (manufacturer_found)
       new_printer_info->desc->manufacturer = manufacturer;
     else
       {
	 int str_len;
	 str_len = strlen(UNKNOWN);
	 manufacturer = malloc(str_len+1);
	 strncpy(manufacturer,UNKNOWN,str_len);
	 manufacturer[str_len]='\0';
	 new_printer_info->desc->manufacturer = manufacturer;
       }
		  
     if (product_found)
       new_printer_info->desc->product = product;   
     else
       {
	 int str_len;

	 str_len = strlen(PRINTER);
     
	 product = malloc(str_len+2);
	 strncpy(product,PRINTER,str_len);
	 sprintf(&product[str_len],"%d",unknown_no);
	 product[str_len+1]='\0';
	 unknown_no++;	  
	 new_printer_info->desc->product = product;
       }

     new_printer_info->desc->dev_no = printer_no;
     new_printer_info->desc->identical = 0;
     new_printer_info->next = NULL;

     /* Insert new record at the end of print info results */
     follow_printer_info_list = &result;
     while(*follow_printer_info_list !=NULL)
       follow_printer_info_list = &((*follow_printer_info_list)->next);

     *follow_printer_info_list = new_printer_info;
	 
     free(device_id);
     close(fd);
     serial_found=0;
     product_found=0;
     manufacturer_found=0;
     count=0;
    }

  return result;
}


extern void search_for_identical_printers(printer_info_t *printer_info_list)
{
  printer_info_t *follow;
    
  follow = printer_info_list;

  while (follow != NULL)
    {
      printer_info_t *old_item;
      printer_info_t *previous;

      int count = 0;
      old_item = follow;
      previous = printer_info_list;
      follow = follow->next;
   
      while (count<old_item->desc->dev_no)
      {
	printer_info_t *prev_item;

	prev_item = previous;
	previous = previous->next;
	if (strcmp(old_item->desc->product,prev_item->desc->product)==0)
	  {
	    old_item->desc->identical++;;
	  }
	count++;
      }
    }
}

 
extern void clear_printer_info(printer_info_t *printer_info_list)
{
	printer_info_t *follow;

	follow = printer_info_list;

	while (follow != NULL)
	{
		printer_info_t *old_item;

		old_item = follow;
		follow = follow->next;
		free(old_item->desc->serial);
		free(old_item->desc->manufacturer);
		free(old_item->desc->product);
		free(old_item->desc);
		free(old_item);
	}
}

extern void display_printer_info(printer_info_t *printer_info_list)
{
  printer_info_t *follow;
    
  follow = printer_info_list;

  while (follow != NULL)
    {
      printer_info_t *old_item;
      old_item = follow;
      follow = follow->next; 
      printf("printer serial# %s\n",old_item->desc->serial);
      printf("printer manufacturer %s\n",old_item->desc->manufacturer);
      printf("printer model %s\n",old_item->desc->product);
      printf("printer is attached to /dev/usb/lp%d\n",old_item->desc->dev_no);
      printf("There are %d identical %s printers already connected\n",
            old_item->desc->identical,old_item->desc->product);
    }
}



extern void store_printer_pool_info(char *pool_name)
{
  char nvram_string[] = "printer_pool_name";

  nvram_set(nvram_string,pool_name);
  nvram_commit();
}

extern char *extract_printer_pool_info()
{
  char *pool_name = nvram_get("printer_pool_name");
  if ((pool_name == NULL) || (pool_name[0] == 0))
      return NULL;
  return(pool_name);
}

extern void install_print_services(pool_info_t *printer_pool)
{
  char *pool_name;
  char buffer[512];

  pool_name = printer_pool->mount_name;
  sprintf(buffer,"/shares/%s/__var",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/locks",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp0",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp1",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp2",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp3",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp4",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp5",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp6",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp7",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp8",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp9",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp10",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp11",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp12",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp13",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp14",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/lp15",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__var/spool/samba",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);

  sprintf(buffer,"/shares/%s/__drivers",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__drivers/W32X86",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__drivers/WIN40",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__drivers/W32ALPHA",pool_name);
  mkdir(buffer, 0700);  
  eval("chmod","0777",buffer);  
  sprintf(buffer,"/shares/%s/__drivers/W32MIPS",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__drivers/W32MIPS",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);
  sprintf(buffer,"/shares/%s/__drivers/W32PPC",pool_name);
  mkdir(buffer, 0700);
  eval("chmod","0777",buffer);


}

extern void uninstall_print_services(pool_info_t *printer_pool)
{
  char buffer[512];

  sprintf(buffer,"/shares/%s/__var/spool",printer_pool->mount_name);
  eval("rm","-rf",buffer);

}

extern void delete_print_services(pool_info_t *printer_pool)
{
  char *pool_name;
  char buffer[512];

  pool_name = printer_pool->mount_name;

  sprintf(buffer,"/shares/%s/__var",pool_name);
  eval("rm","-rf",buffer);

  sprintf(buffer,"/shares/%s/__drivers",pool_name);
  eval("rm","-rf",buffer);

  sprintf(buffer,"/tmp/samba/var/locks/ntdrivers.tdb");
  eval("rm","-f",buffer);

  sprintf(buffer,"/tmp/samba/var/locks/ntprinters.tdb");
  eval("rm","-f",buffer);

  sprintf(buffer,"/tmp/samba/var/locks/ntforms.tdb");
  eval("rm","-f",buffer);

}

extern void move_print_services(pool_info_t *printer_pool,
								pool_info_t *old_printer_pool)
{
  char *pool_name;
  char *old_pool_name;
  char buf1[512];
  char buf2[512];
 
  pool_name = printer_pool->mount_name;
  old_pool_name = old_printer_pool->mount_name;

  sprintf(buf1,"/shares/%s/__var",old_pool_name);
  sprintf(buf2,"/shares/%s/__var",pool_name);
  eval("mv",buf1,buf2);

  sprintf(buf1,"/shares/%s/__drivers",old_pool_name);
  sprintf(buf2,"/shares/%s/__drivers",pool_name);
  eval("mv",buf1,buf2);

  delete_print_services(old_printer_pool);

  return;
}

extern void remove_print_queues(printer_info_t *printer_info_list)
{
  printer_info_t *follow;

  follow = printer_info_list;

  while (follow != NULL)
    {
      printer_info_t *old_item;
      char product_buf[40];

      old_item = follow;
      follow = follow->next;

      if (old_item->desc->identical==0)
	sprintf( product_buf, "%s", old_item->desc->product);
      else
	sprintf( product_buf, "%s%x", 
		 old_item->desc->product,old_item->desc->identical);

      eval("printcmd","queuedel",product_buf);
    }

  eval("killall","printd");
}

extern void setup_printer_tdb_links(pool_info_t *printer_pool)
{
		char *pool_name;
		char path_buffer[1000];
		struct stat status;

		pool_name = printer_pool->mount_name;
		sprintf(path_buffer, "/shares/%s/__var/locks/ntdrivers.tdb",
				pool_name);
		eval("rm", "-f", "/tmp/samba/var/locks/ntdrivers.tdb");
		if (stat(path_buffer,&status)==-1)
		  eval("touch",path_buffer);
		eval("ln", "-s", path_buffer, "/tmp/samba/var/locks/ntdrivers.tdb");

		sprintf(path_buffer, "/shares/%s/__var/locks/ntprinters.tdb",
				pool_name);
		eval("rm", "-f", "/tmp/samba/var/locks/ntprinters.tdb");
		if (stat(path_buffer,&status)==-1)
		  eval("touch",path_buffer);
		eval("ln", "-s", path_buffer, "/tmp/samba/var/locks/ntprinters.tdb");

		sprintf(path_buffer, "/shares/%s/__var/locks/ntforms.tdb",
				pool_name);
		eval("rm", "-f", "/tmp/samba/var/locks/ntforms.tdb");
		if (stat(path_buffer,&status)==-1)
		  eval("touch",path_buffer);
		eval("ln", "-s", path_buffer, "/tmp/samba/var/locks/ntforms.tdb");

}

extern void create_printcap(printer_info_t *printer_info_list,
							pool_info_t *printer_pool)
{
  FILE *fp;
  printer_info_t *follow;

  fp = fopen("/tmp/etc/printcap","w");

  if (fp==NULL)
    {
      fprintf(stderr,
	      "Failed trying write printcap file to /tmp/etc\n");
    } 
	          
  follow = printer_info_list;

  while (follow != NULL)
    {
      printer_info_t *old_item;
      old_item = follow;
      follow = follow->next;

      if (old_item->desc->identical==0)
	{
	  fprintf(fp,
		  "%s|%s %s %s\n"
		  ":sd=/shares/%s/__var/spool/lp%d:\n"
		  ":rp=raw:\n"
		  ":lp=/dev/usb/lp%d\n\n",
		  old_item->desc->product,
		  old_item->desc->manufacturer,
		  old_item->desc->product,
		  old_item->desc->serial,
		  printer_pool->mount_name,
		  old_item->desc->dev_no,
		  old_item->desc->dev_no);
	}
      else
	{
	  fprintf(fp,
		  "%s%x|%s %s %s\n"
		  ":sd=/shares/%s/__var/spool/lp%d:\n"
		  ":rp=raw:\n"
		  ":lp=/dev/usb/lp%d\n\n",
		  old_item->desc->product,
		  old_item->desc->identical,
		  old_item->desc->manufacturer,
		  old_item->desc->product,
		  old_item->desc->serial,
		  printer_pool->mount_name,
		  old_item->desc->dev_no,
		  old_item->desc->dev_no);
	}
    } 

  fclose(fp);

  /* restart smbd */

  eval("killall","-HUP","smbd");
  eval("ln","-sf","/tmp/etc/printcap","/tmp/etc/lpd_printcap");
  eval("ln","-sf","/etc/lpd.conf","/tmp/etc/lpd.conf");
  eval("/bin/printd");
  
  /* send print server commands after printd started */
  for (follow = printer_info_list; follow; follow = follow->next)
    {
      char dev_buf[ 40 ];
      char product_buf[ 40 ];
      const char * argv[ 5 ];
      
      sprintf( dev_buf, "/dev/usb/lp%d", follow->desc->dev_no );

      if (follow->desc->identical==0)
	sprintf( product_buf, "%s", follow->desc->product);
      else
	sprintf( product_buf, "%s%x", 
		 follow->desc->product,follow->desc->identical);
      
      argv[0] = "printcmd";
      argv[1] = "queueadd";
      argv[2] = product_buf;
      argv[3] = dev_buf;
      argv[4] = NULL;
      
      _eval(argv, NULL, 0, NULL );
    } 
}

extern void create_hosts()
{
  FILE *fp;

  fp = fopen("/tmp/hosts","w");

  if (fp==NULL)
    {
      fprintf(stderr,
	      "Failed trying write hosts file to /tmp\n");
    }  

  fprintf(fp,"127.0.0.1	localhost\n");

  fclose(fp);

}

extern void printer_services_init()
{
	char *printer_pool_info;
	printer_info_t *printer_info;

	printer_pool_info = extract_printer_pool_info();
	if (printer_pool_info!=NULL)
	  {
	    all_disk_info_t *all_disk_info;
	    pool_info_t *printer_pool;

	    all_disk_info = read_all_disk_info(POOL_MODE_READ_ONLY);
	    printer_info = get_active_printer_info();
	    search_for_identical_printers(printer_info);
            display_printer_info(printer_info);
	    create_hosts();
	    printer_pool = find_pool(printer_pool_info, all_disk_info);
	    if ((printer_pool != NULL) && (printer_pool->mount_name != NULL))
	      {
	        uninstall_print_services(printer_pool);
	        install_print_services(printer_pool);
		    create_printcap(printer_info,printer_pool);
	      }
	    clear_printer_info(printer_info);  
	    clear_all_disk_info(all_disk_info);
	  }
}


/*
	Search through provided file searching for a record that begins
	with a letter in the accept set without passing over one in
	the reject set.  The reject set may be NULL.
 */
static int
find_record(FILE *fp, const char * accept, const char * reject)
{
	assert(accept != NULL);

	int ch;
	do
	{
		/* Remember where this record started, so we can 
			"unget" the first character. */
		long begin = ftell(fp);
		if (begin == (long) -1)
			return EOF;

		/* Read the first two characters to determine record type. */
		char buf[3]; /* Force to read only first two characters from input. */
		if (fgets(buf, sizeof(buf), fp) == NULL)
			return EOF;

		/* Unget the first two characters. */
		if (fseek(fp, begin, SEEK_SET) == -1)
			return EOF;

		/* Look for a match.  In this case, the file position at
			beginnign of record. */
		if (strspn(buf, accept) > 0 && buf[1] == ':')
		{
			/* We found one of the records that we are lookig for. */
			return buf[0];
		}

		/* Skip rest of this line.  We want to be at next record
			if either this record has an error or if we are
			going to go around the loop again. */
		do
		{
			ch = fgetc(fp);
		} while (ch != '\n' && ch != EOF);


		/* Evaluate the first few characters on this line. */
		if (buf[1] != ':') /* Validate format of line. */
			return EOF; 
		/* See if this is one of the records that should cause us to stop. */
		if (reject != NULL && strspn(buf, reject) > 0)
			return EOF;
	} while (ch == '\n');
	return EOF;
}


/* 
	Find the length of a record from the current position until the next
	newline (or EOF) without affecting our position within the file. 
 */
static int
get_record_len(FILE * fp)
{
	int count = 0;
	long begin;

	begin = ftell(fp);
	if (begin == (long) -1)
		return -1;

	int ch;
	do
	{
		ch = fgetc(fp);
		count++;
	} while (ch != '\n' && ch != EOF);
	if (ch == EOF) /* Include '\n' in count. */
		count--;

	if (fseek(fp, begin, SEEK_SET) == -1)
		return EOF;
	
	return count;
}


/*
	First find the indicated record (see find_record()) and then
	read the record from the file.  The record is read into a 
	dynamically and allocated sized buffer.
 */
static char *
get_record(FILE *fp, const char * accept, const char * reject)
{
	/* Find it's product record. */
	int ch = find_record(fp, accept, reject);
	if (ch == EOF)
		return NULL;
	int len = get_record_len(fp);

	char * line = malloc(len + 1 /* first letter */ + 1 /*null byte*/ + 32 /*safety*/);
	if (line == NULL)
		return NULL;
	if (fgets(line, len+1, fp) == NULL)
	{
		free(line);
		return NULL;
	}
	if (len != strlen(line)) /*excessive caution*/
	{
		free(line);
		return NULL;
	}
	
	return line;
}


static void
remove_newline(char *str)
{
	size_t len = strlen(str);

	if (str[len-1] == '\n')
		str[len-1] = '\0';
}

// There is another instance of this function in nasdock.c.
static const char *
get_separator(void)
{
	char * sep = nvram_safe_get("nasdock_usb_sep");
	if (sep == NULL || sep[0] == '\0')
		sep = "\n\n\n";
	return sep;	
}

static bool
check_class(int class_code)
{
	if (class_code != USB_CLASS_PRINTER && 
		class_code != USB_CLASS_MASS_STORAGE &&
		class_code != USB_CLASS_HUB)
		return FALSE;
	return TRUE;
}

static char * 
get_description(const char * product, bool * p_recognized)
{
	if (product == NULL || product[0] == '\0')
		return NULL;

	/* Parse out device information ids from envrionment variable. */	
	unsigned int vendor_id, product_id, revision_id;
	sscanf((char *)product, "%x/%x/%x", &vendor_id, &product_id, &revision_id);

	/* Experience suggests that I need to cat this file before reading
		it so that the file system updates the contents. */
	const char * dev_file_name = "/proc/bus/usb/devices";
	char buf[512];
	sprintf(buf, "cat %s > /dev/null", dev_file_name);
	system(buf);

	FILE * fp = fopen(dev_file_name, "r");
	if (fp == NULL)
	{
		cprintf("WARNING: Could not open %s\n", dev_file_name);
		goto err;
	}

	/* Search for a Topology record, which begins each unique device. */
	do
	{
		int res;
		char * record;

		/* Find topology record. */
		record = get_record(fp, "T", NULL);
		if (record == NULL)
			goto err;
		free(record); /* Don't really care, just want to consume line. */

		/* Find it's product record. */
		record = get_record(fp, "P", "T");
		char * save = record;
		if (record == NULL)
			goto err;


		/* Skip first part of string record. */
		record += 2; /* skip "P:" */
		while (isspace(*record)) /*skip over white space*/
			record++;

		/* Parse the product record. */
		const char * vendor_format = "Vendor=%x ProdID=%x Rev=%x.%x\n";
			/* copied from linux/drivers/usb/devices.c */

		unsigned int this_vendor, this_product, this_revision_upper, this_revision_lower;
		res = sscanf(record, vendor_format, &this_vendor, &this_product, &this_revision_upper, &this_revision_lower);
		free(save);
		if (res != 4)
			goto err;
		if (((this_revision_upper & ~0xff) != 0) || ((this_revision_lower & ~0xff) != 0))
			goto err;

		unsigned int this_revision = (this_revision_upper << 8) | this_revision_lower;


		/* Is this device the one we are looking for (or effectively identical)? */
		if (	vendor_id == this_vendor && 
			product_id == this_product && 
			revision_id == this_revision )
		{
			/* Iterate over the string records to find the strings we want,
				if they exist. */
			char manu[256 + 32 + 1]; /* USB strings have a max length of 256. */
			char prod[256 + 32 + 1];

			/* Initialize these to empty strings, in case we don't find them. */
			memset(manu, '\0', sizeof(manu));
			memset(prod, '\0', sizeof(prod));
			do
			{
				const char * const manu_prefix = "Manufacturer=";
				const char * const prod_prefix = "Product=";

				/* Now extract the strings for this device. */
				record = get_record(fp, "S", "TCIEBDP"); /* S records should immediately follow. */
				if (record == NULL) /* Normal exit condition. */
					break;
				save = record;

				/* Skip first part of string record. */
				record += 2; /* skip "S:" */
				while (isspace(*record)) /*skip over white space*/
					record++;

				if (strncmp(record, manu_prefix, strlen(manu_prefix)) == 0)
				{
//					record += strlen(manu_prefix);
					strncpy(manu, record, sizeof(manu) - 1);
					remove_newline(manu);
				}
				else if (strncmp(record, prod_prefix, strlen(prod_prefix)) == 0)
				{
//					record += strlen(prod_prefix);
					strncpy(prod, record, sizeof(prod) - 1);
					remove_newline(prod);
				}

				free(save);
			} while (1);

			/*
				For multi-function devices, we only want to report the device
				if we don't recognize any of the devices.
			 */
			*p_recognized = FALSE;
			do
			{
 				/* Now extract the strings for this device. */
				record = get_record(fp, "I", "TBDP"); /* 'T' really should be the end. */
				if (record == NULL) /* Normal exit condition. */
					break;
				save = record;

				/* Skip first part of string record. */
				record += 2; /* skip "I:" */
				while (isspace(*record)) /*skip over white space*/
					record++;

				/* 
					Parse the interface line.  If the parse fails, then 
					assume that the interface is an uncognized class.
				 */
				char * clsstr = strstr(record, "Cls");
				if (clsstr != NULL)
				{
					/* Extract the numeric class identifier. */
					clsstr += 2; /* Skip over 'Cl'. */
					clsstr[0] = '0';
					clsstr[1] = 'x';
					int cls = strtod(clsstr, clsstr+4);

					/* If we recognize the class then record that. */
					if (check_class(cls) == TRUE)
						*p_recognized = TRUE;
				}

				free(save);
			} while (1);
			

			/* We may or may not have strings.  If we do have strings, then build our
				description. */
			int len = strlen(manu) + strlen(prod);
			if (len > 0)
			{
				char * desc = malloc(len + 5 + 1/*null byte*/ + 32 /*safety*/);
				if (desc == NULL)
					goto err;

				strcpy(desc, manu);
				if (manu[0] != '\0')
				{
					const char * sep = get_separator();
					strcat(desc, sep);
				}
				strcat(desc, prod);

				fclose(fp);
				return desc;
			}

			goto err; /* no strings */
		}
	} while (1);

err:
	fclose(fp);
	return NULL;
}


static bool
check_printer(void)
{
	/* Check to be sure that this is a printer.  GEW*/
	char * action = getenv("ACTION");
	char * type = getenv("TYPE");
	char * product = getenv("PRODUCT");
	char * device = getenv("DEVICE");
	if (	action != NULL &&
		strcmp(action, "add") == 0 &&
		type != NULL) 
		/* Don't care if product is NULL. */
	{
		/* The class code may be either in the TYPE variable or the
			INTERFACE variable. */
		int class_code = atoi(type);
		if (class_code == USB_CLASS_PER_INTERFACE)
		{
			char * iface = getenv("INTERFACE");
			if (iface != NULL)
				class_code = atoi(iface);
		}
		if (check_class(class_code) == FALSE)
		{
			int dev = 0; /* 0 means unknown */
			if (device != NULL)
				sscanf(device, "/proc/bus/usb/%03d/", &dev);

			char event_name[32];
			sprintf(event_name, "usb%d", dev);

			bool recognized = FALSE;
			char * desc = get_description(product, &recognized);
			if (recognized == TRUE)
				return TRUE;

			/* Send a message via nasdock/naspp. */
			if (desc == NULL)
				ipc_send_msg("nasdock", event_name, "Unrecognized USB device"); // Ignore return code.
			else
			{
				ipc_send_msg("nasdock", event_name, desc); // Ignore return code.
			}

			/* Send a user notification. */
			char summary[128];
			char key[80];
			char details[3000];
			char detail_dev[80];
			char detail_desc[1024];

			if (dev == 1 || dev == 2)
			{
				sprintf(summary, TRANSLATE("Unrecognized USB device in port %d."), dev);
				detail_dev[0] = '\0';
			}
			else
			{
				sprintf(summary, TRANSLATE("Unrecognized USB device into a USB port."));
				sprintf(detail_dev, TRANSLATE("The USB device is on port %d.\n\n"), dev);
			}
			sprintf(key, "unknown_USB_dev_in_port_%d", dev);

			char * detail_base = TRANSLATE(
"The BroadNAS has detected a USB device that may not be a supported\n" 
"device type.  If the USB device is neither a disk drive nor a printer, \n"
"you should unplug it.\n\n"
);
			if (desc == NULL)
			{
				detail_desc[0] = '\0';
			}
			else
			{
				const char * sep = get_separator();
				char * prod = strstr(desc, sep);
				*prod = '\0'; /*terminate manufacturer description*/
				prod += strlen(sep); /* Skip over separator. */

				if (prod == NULL)
				{
					/* Cautiously protect ourselves since we have fixed length buffers. */
					/* USB strings should have a max length of 256 anyhow. */
					if (strlen(desc) > 512)
						desc[513] = '\0';
					
					sprintf(detail_desc, TRANSLATE(
"The device description is as follows:\n\n\t%s"), desc);
				}
				else
				{
					/* Cautiously protect ourselves since we have fixed length buffers. */
					/* USB strings should have a max length of 256 anyhow. */
					if (strlen(desc) > 256)
						desc[257] = '\0';
					if (strlen(prod) > 256)
						prod[257] = '\0';
					
					sprintf(detail_desc, TRANSLATE(
"The device is a '%s' made by '%s'.\n\n"), prod, desc);
				}
				free(desc);
			}

			sprintf(details, "%s%s%s", detail_base, detail_dev, detail_desc);
			create_user_note(	summary, key, NULL /*link*/, details);

			return FALSE; /*it is not a printer*/
		}
		/* Else, got a printer, so mount it. */
	}
	/* Else, we have no clue what is going on.  */

	return TRUE; /*declare it a printer*/
}


extern int hotplug_printer()
{
	/* We ignore the return code, because we try to mount even non-printers as
		printers.  This is helpful for printer devices that don't use the USB 
		printer type. */
	check_printer();

	char *printer_pool_info;
	printer_info_t *printer_info;

	printer_pool_info = extract_printer_pool_info();
	if (printer_pool_info!=NULL)
	  {
	    all_disk_info_t *all_disk_info;
	    pool_info_t *printer_pool;

	    all_disk_info = read_all_disk_info(POOL_MODE_READ_ONLY);
	    printer_info = get_active_printer_info();
	    search_for_identical_printers(printer_info);
	    create_hosts();
	    printer_pool = find_pool(printer_pool_info, all_disk_info);
	    create_printcap(printer_info,printer_pool);
	    clear_printer_info(printer_info);  
	    clear_all_disk_info(all_disk_info);
	  }

	if (strcmp(nvram_safe_get("cifs_master_enable"), "enabled") == 0)
	{
		/* Tell rc to update smb.conf. */
		notify_rc("share_update");

		/* Give rc time to update smb.conf. */
		sleep(5);

		/* Tell smbd to re-read its configuration file. */
		eval("killall", "-HUP", "smbd");
	}

	return 0;
}
